#!/usr/bin/env python3

# standards
from typing import Dict, Iterable, List, Optional, Tuple, Union, no_type_check
from urllib.parse import urljoin

# 3rd parties
from requests import Request

# klon
from .utils import Element


def parse_form(form: Element, base_url: Optional[str] = None) -> Request:
    if not form.tag == 'form':
        raise ValueError(f'Expected <form> node, got <{form.tag}>')
    req = Request(
        method=form.get('method', 'GET'),
        url=_parse_form_action(form, base_url),
    )
    data = _parse_form_data(form)
    if req.method == 'GET':
        req.params = data
    else:
        req.data = data
    return req


def _parse_form_action(form: Element, base_url: Optional[str]) -> str:
    action = form.get('action')
    if base_url:
        if action:
            action = urljoin(base_url, action)
        else:
            action = base_url
    elif not action:
        raise ValueError('Form has no action attribute, and no `base_url` was given')
    return action


@no_type_check
def _parse_form_data(form: Element) -> Dict[str, str]:
    data: Dict[str, Union[str, List[str]]] = {}
    for itype, name, value in _parse_input_name_value_pairs(form):
        if name in data and itype != 'radio':
            if not isinstance(data[name], list):
                data[name] = [data[name]]
            data[name].append(value)
        else:
            data[name] = value
    return data


def _parse_input_name_value_pairs(form: Element) -> Iterable[Tuple[str, str, str]]:
    for node in form.iter():
        name = node.get('name')
        if not name:
            pass
        elif node.tag == 'input':
            value = node.get('value')
            itype = node.get('type', '')
            if itype in ('checkbox', 'radio'):
                if node.get('checked'):
                    yield itype, name, value or 'on'
            elif itype in ('submit', 'button', 'image', 'reset'):
                # We ignore these completely; if the user wants to simulate a click, the field has to be set manually on the
                # returned Request's `data` dict
                pass
            else:
                # Text etc inputs without a value get the empty string
                yield itype, name, value or ''
        elif node.tag == 'select':
            valued_options: List[Element] = list(node.xpath('.//option[@value]'))  # type: ignore
            selected_options: List[Element] = list(node.xpath('.//option[@selected]'))  # type: ignore
            if node.get('multiple'):
                for option in selected_options:
                    yield 'select', name, option.get('value', '')
            elif selected_options:
                yield 'select', name, selected_options[-1].get('value', '')
            elif valued_options:
                yield 'select', name, valued_options[0].get('value')  # type: ignore



#!/usr/bin/env python3

# standards
import re
from typing import no_type_check
from urllib.parse import urljoin

# 3rd parties
import lxml.etree as ET


TAGS_WITH_URL_ATTRIBUTES = {
    'a': ['href'],
    'area': ['href'],
    'audio': ['src'],
    'base': ['href'],
    'blockquote': ['cite'],
    'del': ['cite'],
    'embed': ['src'],
    'form': ['action'],
    'frame': ['src'],
    'iframe': ['src'],
    'img': ['src', 'srcset'],
    'input': ['src'],
    'ins': ['cite'],
    'link': ['href'],
    'object': ['data'],
    'script': ['src'],
    'source': ['src', 'srcset'],
    'track': ['src'],
    'video': ['poster', 'src'],
}


XPATH_TAGS_WITH_URL_ATTRIBUTES = '//*[%s]' % ' or '.join(
    f'self::{tag}' for tag in sorted(TAGS_WITH_URL_ATTRIBUTES)
)


def parse_html_etree(html_str: str, remove_comments: bool = False) -> ET._Element:
    if not isinstance(html_str, str):
        raise TypeError('Expected str, not %s; %s' % (type(html_str).__name__, repr(html_str)[:100]))

    # It's rare for HTML docs to contain an XML declaration, but when they do, lxml throws an exception, so remove them
    html_str = re.sub(r'^<\?xml[^>]+\?>', '', html_str)

    html_str = html_str.strip()
    if html_str == '':
        raise ValueError("Can't parse HTML etree from an empty string")

    parser = ET.HTMLParser(remove_comments=remove_comments)
    return ET.HTML(html_str, parser)


@no_type_check  # until lxml-stubs improves
def make_all_urls_absolute(base_url: str, etree: ET._Element) -> None:
    """
    Modify all links in the given HTML etree to be absolute URLs, using the given `base_url` to resolve relative URLs.
    """
    for node in etree.xpath(XPATH_TAGS_WITH_URL_ATTRIBUTES):  # type: ignore[union-attr]
        for attr in TAGS_WITH_URL_ATTRIBUTES[node.tag]:
            value = node.get(attr)
            if value is not None:
                if attr == 'srcset':
                    new_value = re.sub(
                        r'((?:^|,)\s*)(\S+)',
                        lambda m: m[1] + urljoin(base_url, m[2]),
                        value,
                    )
                else:
                    new_value = urljoin(base_url, value)
                if new_value != value:
                    node.set(attr, new_value)


@no_type_check
def extract_js_str(element: ET._Element) -> str:
    return '\n\n'.join(
        re.sub(r'^\s*<!--', '', re.sub(r'-->\s*$', '', script_str))
        for script_str in element.xpath('.//script/text()')
    )



#!/usr/bin/env python3

# standards
from pathlib import Path
import re

# 3rd parties
import pytest

# klon
from klon import parse_form, parse_html_etree


def _load_fixture(method):
    html_file = Path(__file__).parent / 'test_forms.html'
    doc = parse_html_etree(html_file.read_text('UTF-8'))
    form = doc.xpath('//form')[0]
    if method is None:
        form.attrib.pop('method', '')
    else:
        form.attrib['method'] = method
    expected_query_string = doc.xpath('//a/@href')[0].lstrip('?')
    return form, expected_query_string


@pytest.mark.parametrize('method', [None, 'GET', 'POST'])
def test_form_parser(method):
    form, expected_query_string = _load_fixture(method)
    obtained_preq = parse_form(form).prepare()
    obtained_qs = re.sub(r'.+\?', '', obtained_preq.url) if method in (None, 'GET') else obtained_preq.body
    assert obtained_qs == expected_query_string


@pytest.mark.parametrize('method', [None, 'GET', 'POST'])
def test_form_parser_method(method):
    form, _qs_unused = _load_fixture(method)
    assert parse_form(form).method == method or 'GET'



#!/usr/bin/env python3

# standards
from typing import Union

# 3rd parties
import lxml.etree as ET


def parse_xml_etree(xml_str: Union[bytes, str], remove_comments: bool = False) -> ET._Element:
    parser = ET.XMLParser(remove_comments=remove_comments)
    return ET.XML(xml_str, parser)



#!/usr/bin/env python3

# 3rd parties
import pytest

# klon
from klon import Element, build_etree, make_all_urls_absolute, parse_html_etree, tostring


@pytest.mark.parametrize(
    'base_url, etree, expected',
    [
        (
            # relative URLs are made absolute
            'http://example.com/',
            build_etree('div', ['a', {'href': 'relative/link'}]),
            build_etree('div', ['a', {'href': 'http://example.com/relative/link'}]),
        ),
        (
            # absolute URLs are untouched
            'http://example.com/',
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
        ),
        (
            # an https:// URL doesn't override an http:// URL
            'https://example.com/',
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
        ),
        (
            # an #anchor link gets merged too
            'http://example.com/',
            build_etree('div', ['a', {'href': '#anchor'}]),
            build_etree('div', ['a', {'href': 'http://example.com/#anchor'}]),
        ),
        (
            # not just ahrefs are updated -- see full list in klon/html.py
            'http://example.com/',
            build_etree('div', ['img', {'src': 'image.jpg'}]),
            build_etree('div', ['img', {'src': 'http://example.com/image.jpg'}]),
        ),
        (
            # the root element is updated too
            'http://example.com/',
            build_etree('a', {'href': 'relative/link'}),
            build_etree('a', {'href': 'http://example.com/relative/link'}),
        ),
        (
            # if the URL attribute is missing we don't complain
            'http://example.com/',
            build_etree('div', ['a', {'name': 'rover'}]),
            build_etree('div', ['a', {'name': 'rover'}]),
        ),
        (
            # empty links are replaced by the full base URL, as a browser does
            'http://example.com/',
            build_etree('div', ['a', {'href': ''}]),
            build_etree('div', ['a', {'href': 'http://example.com/'}]),
        ),
    ]
)
def test_make_all_urls_absolute(base_url, etree, expected):
    make_all_urls_absolute(base_url, etree)
    assert tostring(etree) == tostring(expected)


@pytest.mark.parametrize(
    'input_str, expected_tree',
    [
        # We're not going to test the intricacies of lxml's HTML-parsing abilities here (e.g. how it handles mismatched angle
        # brackets), but rather just test the behaviour of our function.
        (
            '<html><body><h1>Hello</h1></body></html>',
            ['html', ['body', ['h1', 'Hello']]],
        ),
        (
            # <html><body> is auto-inserted (this is done by lxml)
            '<h1>Hello</h1>',
            ['html', ['body', ['h1', 'Hello']]],
        ),
        (
            # a plain string is converted to a <p> (this is done by lxml)
            'Hello',
            ['html', ['body', ['p', 'Hello']]],
        ),
        (
            # a None input triggers a TypeError
            None,
            TypeError,
        ),
        (
            # an empty string triggers a ValueError
            '',
            ValueError,
        ),
        (
            # a string consisting of whitespace only also triggers a ValueError
            ' ',
            ValueError,
        ),
        (
            # a string consisting of an xml declaration only counts as an empty string
            '<?xml version="1.0" encoding="UTF-8" ?>',
            ValueError,
        ),
    ]
)
def test_html_parser(input_str, expected_tree):

    def compare(obtained: Element, expected: Element) -> None:
        assert obtained.tag == expected.tag
        for key, value in expected.attrib.items():
            assert obtained.get(key) == value
        assert len(obtained) == len(expected)
        for ochild, echild in zip(obtained, expected):
            compare(ochild, echild)

    if isinstance(expected_tree, type) and issubclass(expected_tree, Exception):
        with pytest.raises(expected_tree):
            parse_html_etree(input_str)
    else:
        obtained = parse_html_etree(input_str)
        expected = build_etree(*expected_tree)
        compare(obtained, expected)



#!/usr/bin/env python3

# 3rd parties
import pytest

# klon
from klon import extract_text, parse_html_etree


@pytest.mark.parametrize(
    'html, expected_default, expected_multiline',
    [
        (
            '<p>This is a   <b>test!</b></p>   And it works.',
            'This is a test! And it works.',
            'This is a test!\n\nAnd it works.',
        ),
        (
            '<ul><li>one<li>two</ul>',
            'one two',
            'one\n\ntwo',
        ),
        (
            '''
              <p>One paragraph</p>
              <p>two paragraphs</p>
              three and<br>
              four
              <p>and five<br></p>
              and six and<br>
              <p>seven</p>
            ''',
            'One paragraph two paragraphs three and four and five and six and seven',
            'One paragraph\n\ntwo paragraphs\n\nthree and\nfour\n\nand five\n\nand six and\n\nseven',
        ),
        (
            '''
            <html>
              <head><title><Test</title></head>
              <body>
                <script> alert("Boo!") </script>
                <p>Dr Livingstone, I presume?</p>
              </body>
            </html>
            ''',
            'Dr Livingstone, I presume?',
            'Dr Livingstone, I presume?',
        ),
        (
            '<p>It was <!-- not --> a dark stormy night</p>',
            'It was a dark stormy night',
            'It was a dark stormy night',
        ),
        (
            '<p>It was <style> p { font-weight: bold; } </style> a dark stormy night</p>',
            'It was a dark stormy night',
            'It was a dark stormy night',
        ),
        (
            '<p>It was a <span id="root">dark stormy</span> night</p>',
            'dark stormy',
            'dark stormy',
        ),
    ]
)
def test_extract_text(html, expected_default, expected_multiline):
    etree = parse_html_etree(html)
    roots = etree.xpath('//*[@id="root"]')
    if roots:
        etree, = roots
    assert extract_text(etree) == expected_default
    assert extract_text(etree, multiline=True) == expected_multiline
